@@ -74,7 +74,7 @@ extern int i386_debug;
 
 union YYSTYPE
 {
-#line 217 "i386_parse.y" /* yacc.c:1909  */
+#line 216 "/home/mark/src/elfutils/libcpu/i386_parse.y" /* yacc.c:1909  */
 
   unsigned long int num;
   char *str;
