@@ -23,7 +23,6 @@
 #include <argp.h>
 #include <assert.h>
 #include <errno.h>
-#include <error.h>
 #include <fcntl.h>
 #include <gelf.h>
 #include <libdw.h>
@@ -36,8 +35,8 @@
 #include <string.h>
 #include <unistd.h>
 
-#include <system.h>
-
+#include <printversion.h>
+#include "system.h"
 
 struct segments
 {
